//================================================================================================================================
//
//  Copyright (c) 2015-2018 VisionStar Information Technology (Shanghai) Co., Ltd. All Rights Reserved.
//  EasyAR is the registered trademark or trademark of VisionStar Information Technology (Shanghai) Co., Ltd in China
//  and other countries for the augmented reality technology developed by VisionStar Information Technology (Shanghai) Co., Ltd.
//
//================================================================================================================================

package cn.easyar.samples.helloarrecording;

import android.opengl.GLES20;
import android.util.Log;

import cn.easyar.*;

public class HelloAR
{
    private CameraDevice camera;
    private CameraFrameStreamer streamer;
    private Renderer videobg_renderer;
    private RecorderRenderer recorder_renderer;
    private Recorder recorder;
    private boolean viewport_changed = false;
    private Vec2I view_size = new Vec2I(0, 0);
    private int rotation = 0;
    private Vec4I viewport = new Vec4I(0, 0, 1280, 720);
    private boolean recording_started = false;

    public boolean initialize()
    {
        camera = new CameraDevice();
        streamer = new CameraFrameStreamer();
        streamer.attachCamera(camera);

        boolean status = true;
        status &= camera.open(CameraDeviceType.Default);
        camera.setSize(new Vec2I(1280, 720));
        return status;
    }

    public void dispose()
    {
        if (recorder != null) {
            recorder.dispose();
            recorder = null;
        }
        recorder_renderer = null;
        if (videobg_renderer != null) {
            videobg_renderer.dispose();
            videobg_renderer = null;
        }
        if (streamer != null) {
            streamer.dispose();
            streamer = null;
        }
        if (camera != null) {
            camera.dispose();
            camera = null;
        }
    }

    public boolean start()
    {
        boolean status = true;
        status &= (camera != null) && camera.start();
        status &= (streamer != null) && streamer.start();
        camera.setFocusMode(CameraDeviceFocusMode.Continousauto);
        return status;
    }

    public boolean stop()
    {
        boolean status = true;
        status &= (streamer != null) && streamer.stop();
        status &= (camera != null) && camera.stop();
        return status;
    }

    public void initGL()
    {
        if (videobg_renderer != null) {
            videobg_renderer.dispose();
            videobg_renderer = null;
        }
        if (recorder != null) {
            recorder.dispose();
            recorder = null;
        }
        videobg_renderer = new Renderer();
        recorder_renderer = new RecorderRenderer();
        recorder = new Recorder();
    }

    public void resizeGL(int width, int height)
    {
        view_size = new Vec2I(width, height);
        viewport_changed = true;
    }

    private void updateViewport()
    {
        CameraCalibration calib = camera != null ? camera.cameraCalibration() : null;
        int rotation = calib != null ? calib.rotation() : 0;
        if (rotation != this.rotation) {
            this.rotation = rotation;
            viewport_changed = true;
        }
        if (viewport_changed) {
            Vec2I size = new Vec2I(1, 1);
            if ((camera != null) && camera.isOpened()) {
                size = camera.size();
            }
            if (rotation == 90 || rotation == 270) {
                size = new Vec2I(size.data[1], size.data[0]);
            }
            float scaleRatio = Math.max((float) view_size.data[0] / (float) size.data[0], (float) view_size.data[1] / (float) size.data[1]);
            Vec2I viewport_size = new Vec2I(Math.round(size.data[0] * scaleRatio), Math.round(size.data[1] * scaleRatio));
            viewport = new Vec4I((view_size.data[0] - viewport_size.data[0]) / 2, (view_size.data[1] - viewport_size.data[1]) / 2, viewport_size.data[0], viewport_size.data[1]);

            if ((camera != null) && camera.isOpened())
                viewport_changed = false;

            if (recording_started) {
                recorder_renderer.resize(view_size.data[0], view_size.data[1]);
            }
        }
    }

    public void preRender()
    {
        if (!recording_started) {
            return;
        }
        recorder_renderer.preRender();
    }

    public void render()
    {
        GLES20.glClearColor(1.f, 1.f, 1.f, 1.f);
        GLES20.glClear(GLES20.GL_COLOR_BUFFER_BIT | GLES20.GL_DEPTH_BUFFER_BIT);

        if (videobg_renderer != null) {
            Vec4I default_viewport = new Vec4I(0, 0, view_size.data[0], view_size.data[1]);
            GLES20.glViewport(default_viewport.data[0], default_viewport.data[1], default_viewport.data[2], default_viewport.data[3]);
            if (videobg_renderer.renderErrorMessage(default_viewport)) {
                return;
            }
        }

        if (streamer == null) { return; }
        Frame frame = streamer.peek();
        try {
            updateViewport();
            GLES20.glViewport(viewport.data[0], viewport.data[1], viewport.data[2], viewport.data[3]);

            if (videobg_renderer != null) {
                videobg_renderer.render(frame, viewport);
            }
        }
        finally {
            frame.dispose();
        }
    }

    public void postRender()
    {
        if (!recording_started) {
            return;
        }
        recorder_renderer.postRender(viewport);
        recorder.updateFrame();
    }

    public void requestPermissions(FunctorOfVoidFromPermissionStatusAndString callback)
    {
        recorder.requestPermissions(callback);
    }

    public void startRecording(String path, final FunctorOfVoidFromRecordStatusAndString callback)
    {
        if (recording_started) { return; }
        recorder.setOutputFile(path);
        recorder.setZoomMode(RecordZoomMode.ZoomInWithAllContent);
        recorder.setProfile(RecordProfile.Quality_720P_Middle);
        recorder_renderer.resize(view_size.data[0], view_size.data[1]);
        recorder.setInputTexture(recorder_renderer.getTextureId(), view_size.data[0], view_size.data[1]);
        int mode = view_size.data[0] < view_size.data[1] ? RecordVideoOrientation.Portrait:RecordVideoOrientation.Landscape;
        recorder.setVideoOrientation(mode);
        recorder.open(new FunctorOfVoidFromRecordStatusAndString() {
            @Override
            public void invoke(int status, String value) {
                if (status == RecordStatus.OnStopped) {
                    recording_started = false;
                }
                callback.invoke(status, value);
            }
        });
        recorder.start();
        recording_started = true;
    }

    public void stopRecording()
    {
        if (!recording_started) { return; }
        recorder.stop();
        recording_started = false;
    }
}
